
/* See LICENSE file for copyright and license details. */
#include <getopt.h>
#include <magic.h>
#include <regex.h>
#include <spawn.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define _POSIX_C_SOURCE 200809L

#define REGEX_PROTOCOL "^[a-z]*://"
#define REGEX_URL      "^(([a-z]*)://(.*)/)?(.*)$"

#define USAGE "usage: sopen [-h] [-v] [-d] <url/file>"

#define STR_LITERAL(x) #x
#define STRINGIFY(x) STR_LITERAL(x)

typedef struct {
	const char     *protocol_regex;
	const char     *url_regex;
	const char     *mime_regex;
	const char     *file_regex;
	const char     *program;
} Rule;

void            print_help();
void            print_version();
void            parse_args(const int argc, char *const *argv, int *dry_run,
                           int *url_ind);
int             parse_url(const char *url, char *protocol, char *file);
int             match_rule(const Rule rule, const char *url,
                           const char *protocol, const char *file,
                           const char *mime_type);

#include "config.h"

void
print_help() {
	printf(USAGE "\n");
}

void
print_version() {
	printf(STRINGIFY(PROGNAME) " " STRINGIFY(VERSION) "\n");
	printf("CFLAGS:  " STRINGIFY(CFLAGS) "\n");
	printf("LDFLAGS: " STRINGIFY(LDFLAGS) "\n");
}

void
parse_args(const int argc, char *const *argv, int *dry_run, int *url_ind) {

	char           *cvalue = NULL;
	int             index;
	int             c;

	opterr = 0;

	while((c = getopt(argc, argv, "dhV")) != -1) {
		switch (c) {
		case 'd':
			*dry_run = 1;
			break;
		case 'h':
			print_help();
			exit(0);
		case 'V':
			print_version();
			exit(0);
		case '?':
			fprintf(stderr,
			        "Unknown option `\\x%x'. Run `sopen -h' for usage.\n",
			        optopt);
			exit(1);
		default:
			exit(255);
		}
	}
	*url_ind = optind;
}

int
parse_url(const char *url, char *protocol, char *file) {

	regex_t         regex_url;
	regmatch_t      match[8];
	int             res;

	res = regcomp(&regex_url, REGEX_URL, REG_EXTENDED);
	if(res > 0)
		return 1;

	res = regexec(&regex_url, url, regex_url.re_nsub + 1, match, 0);
	if(res > 0) {
		printf("Failed to parse url: %s\n", url);
		return 1;
	}

	memcpy(protocol, url + match[2].rm_so, match[2].rm_eo - match[2].rm_so);
	protocol[match[2].rm_eo - match[2].rm_so] = '\0';

	memcpy(file, url + match[4].rm_so, match[4].rm_eo - match[4].rm_so);
	file[match[4].rm_eo - match[4].rm_so] = '\0';

	return 0;
}

int
match_rule(const Rule rule, const char *url, const char *protocol,
	   const char *file, const char *mime_type) {

	regex_t         regex;
	int             res;

	int             dry_run = 0;

	res = regcomp(&regex, rule.url_regex, REG_EXTENDED);
	if(res > 0) {
		printf("Failed to compile url regex: %s\n", rule.url_regex);
		return 1;
	}
	res = regexec(&regex, url, 0, NULL, 0);
	if(res > 0)
		return 1;

	res = regcomp(&regex, rule.protocol_regex, REG_EXTENDED);
	if(res > 0) {
		printf("Failed to compile protocol regex: %s\n",
		       rule.protocol_regex);
		return 1;
	}
	res = regexec(&regex, protocol, 0, NULL, 0);
	if(res > 0)
		return 1;

	res = regcomp(&regex, rule.file_regex, REG_EXTENDED);
	if(res > 0) {
		printf("Failed to compile file regex: %s\n", rule.file_regex);
		return 1;
	}
	res = regexec(&regex, file, 0, NULL, 0);
	if(res > 0)
		return 1;

	res = regcomp(&regex, rule.mime_regex, REG_EXTENDED);
	if(res > 0) {
		printf("Failed to compile mime regex: %s\n", rule.mime_regex);
		return 1;
	}
	res = regexec(&regex, mime_type, 0, NULL, 0);
	if(res > 0)
		return 1;

	return 0;
}

int
main(int argc, char *const *argv) {

	int             res;
	int             dry_run;
	int             url_ind;

	magic_t         cookie;

	regex_t         has_protocol;

	const char     *mime_type;

	const char     *url;
	char            protocol[16];
	char            file[1024];

	char           *child_args[3];

	cookie = magic_open(MAGIC_MIME_TYPE);
	if(cookie == NULL) {
		printf("Failed to open magic cookie.\n");
		goto error;
	}
	magic_load(cookie, NULL);

	parse_args(argc, argv, &dry_run, &url_ind);
	url = argv[url_ind];

	printf("%s\n", url);
	printf("%d\n", dry_run);

	res = regcomp(&has_protocol, REGEX_PROTOCOL, REG_EXTENDED);
	if(res > 0)
		goto error;

	mime_type = "";

	res = regexec(&has_protocol, url, 0, NULL, 0);
	if(res > 0)
		mime_type = magic_file(cookie, url);

	for(int n = 0; n < sizeof(rules) / sizeof(Rule); n++) {

		res = parse_url(url, &protocol[0], &file[0]);
		if(res > 0)
			continue;

		res = match_rule(rules[n], url, protocol, file, mime_type);

		if(res == 0) {

			child_args[0] = (char *)rules[n].program;
			child_args[1] = (char *)url;
			child_args[2] = NULL;
			if(dry_run) {
				printf("%s %s\n", child_args[0], child_args[1]);
			} else {
				execvp(rules[n].program, child_args);
			}
			return 0;
		}
	}

	error:
	magic_close(cookie);
	return 1;
}
